package org.telosys.tools.repository.model;

import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.Hashtable;
import java.util.Iterator;

public class RepositoryModel 
{

	private String version ;
	
	private String databaseName ;

	private String databaseType ;
	
	private Date generationDate ;
	
	private Date lastUpdateDate ;
	
	private Hashtable<String,Entity> htEntities = new Hashtable<String,Entity>() ; 

	public String getVersion() {
		return version;
	}

	public void setVersion(String version) {
		this.version = version;
	}

	public String getDatabaseName() {
		return databaseName;
	}

	public void setDatabaseName(String databaseName) {
		this.databaseName = databaseName;
	}

	public String getDatabaseType() {
		return databaseType;
	}

	public void setDatabaseType(String databaseType) {
		this.databaseType = databaseType;
	}

	public Date getGenerationDate() {
		return generationDate;
	}

	public void setGenerationDate(Date generationDate) {
		this.generationDate = generationDate;
	}

	public Date getLastUpdateDate() {
		return lastUpdateDate;
	}

	public void setLastUpdateDate(Date lastUpdateDate) {
		this.lastUpdateDate = lastUpdateDate;
	}
	
	//-------------------------------------------------------------------------------
	
	public int getNumberOfEntities() {
		return htEntities.size();
	}

	/**
	 * Returns an array of all the entities of the model.<br>
	 * The entities are sorted by name.
	 * 
	 * @return
	 */
	public Entity[] getEntities() {
		Entity[] array = (Entity[]) htEntities.values().toArray( new Entity[htEntities.size()] );
		Arrays.sort(array);
		return array ;
	}
	
	/**
	 * Returns a collection of all the entities of the model.<br>
	 * The entities are sorted by name.
	 * 
	 * @return
	 */
	public Collection getEntitiesCollection() {
		//return htEntities.values() ;
		Entity[] entities = getEntities();
		return Arrays.asList(entities);
	}
	
	/**
	 * Returns an array containing the names of all the entities of the model.<br>
	 * The names are sorted in alphabetic order.
	 * 
	 * @return
	 */
	public String[] getEntitiesNames() {
		Collection values = htEntities.values();
		String[] names = new String[values.size()];
		int cpt = 0;
		for (Iterator iterator = values.iterator(); iterator.hasNext();) {
			Entity entity = (Entity) iterator.next();
			names[cpt] = entity.getName();
			cpt++;
		}
		Arrays.sort(names);
		return names;
	}
	
	public Entity getEntityByName(String name) {
		return (Entity) htEntities.get(name);
	}
	
	public void storeEntity(Entity entity) {
		htEntities.put(entity.getName(), entity);
	}
	
	public void removeEntity(String name) {
		htEntities.remove(name);
	}
	
	//-------------------------------------------------------------------------------
	// LINKS management
	//-------------------------------------------------------------------------------
	/**
	 * Returns the link for the given id
	 * @param id
	 * @return
	 */
	public Link getLinkById(String id) {
		if ( id != null ) {
			Entity [] entities = this.getEntities();
			for ( int i = 0 ; i < entities.length ; i++ ) {
				Entity entity = entities[i];
				Link [] links = entity.getLinks();
				for ( int j = 0 ; j < links.length ; j++ ) {
					Link link = links[j];
					if ( id.equals( link.getId() ) )  {
						return link;
					}
				}
			}
		}
		return null ;
	}

	//-------------------------------------------------------------------------------
	/**
	 * Removes all the links 
	 */
	public void removeAllLinks() {
		Entity [] entities = this.getEntities();
		for ( int i = 0 ; i < entities.length ; i++ ) {
			Entity entity = entities[i];
			entity.removeAllLinks();
		}
	}
	
	//-------------------------------------------------------------------------------
	/**
	 * Removes the link corresponding to the given id
	 * @param id
	 */
	public void removeLinkById(String id) {
		Link link = getLinkById(id);
		if ( link != null ) {
			//--- Remove link 
			Entity entity = getEntityByName( link.getSourceTableName() );
			if ( entity != null ) {
				entity.removeLink(link);
			}
		}
	}
	//-------------------------------------------------------------------------------
	/**
	 * Returns the RelationLinks ( the 2 links of a relation ) for the given link id
	 * @param linkId the id of one of the 2 links of the relation
	 * @return
	 */
	public RelationLinks getRelationByLinkId(String linkId) 
	{
		Link link1 = getLinkById(linkId);
		if ( link1 != null ) {
			if ( link1.isOwningSide() ) {
				//--- Owning Side => try to found the inverse side
				Entity [] entities = this.getEntities();
				for ( int i = 0 ; i < entities.length ; i++ ) {
					Entity entity = entities[i];
					Link [] links = entity.getLinks();
					for ( int j = 0 ; j < links.length ; j++ ) {
						Link link2 = links[j];
						if ( link2.isOwningSide() == false ) {
							if ( linkId.equals( link2.getInverseSideOf() ) ) {
								return new RelationLinks ( link1, link2 );
							}
						}
					}
				}
				// inverse side not found 
				return new RelationLinks ( link1, null );
			}
			else {
				//--- Inverse Side => try to found the owning side
				Link link2 = getLinkById( link1.getInverseSideOf() ) ;
				return new RelationLinks ( link2, link1 );
			}
		}
		return null ;
	}
	
	//-------------------------------------------------------------------------------
	// FOREIGN KEYS management
	//-------------------------------------------------------------------------------
	/**
	 * Search and return a Foreign Key  
	 * @param fkName the name to e searched
	 * @return the Foreign Key or null if not found
	 */
	public ForeignKey getForeignKeyByName(String fkName)
	{
		Entity [] entities = this.getEntities();
		for ( Entity entity : entities ) {
			ForeignKey fk = entity.getForeignKey(fkName);
			if ( fk != null ) {
				return fk ; // FOUND 
			}
		}
		return null ;
	}
	
}
